using System.Collections.ObjectModel;
using System.Windows.Forms;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace EQIA
{
    /// <summary>
    /// This class serves to provide a link between the xna drawing and a control, it will tamper with the back buffer
    /// to create the desired effect but overall works to allow xna to draw on top of multiple controls at once. Since
    /// this class will clear the part of the back buffer that the control sits on top of, it's advised to draw
    /// these last.
    /// </summary>
    public class ControlSurface
    {
        #region Fields

        private Game myGame;
        private Control myControl;
        private SpriteBatch mySpriteBatch;

        // Surface
        private RenderTarget2D myRenderTarget;
        private Texture2D myTexture;
        private Viewport myViewport;

        // Background
        private Texture2D myBackTexture;
        private Color myBackColor;
        private string myBackTexturePath;

        // Tracking
        private bool isTextureReload;
        private bool isRequireRedraw;
        private bool isVisible;

        #endregion

        #region Properties

        /// <summary>
        /// The game that this control surface belongs to.
        /// </summary>
        public Game Game { get { return myGame; } }

        /// <summary>
        /// The graphics device this control surface will use to draw it's contents.
        /// </summary>
        public GraphicsDevice GraphicsDevice { get { return myGame.GraphicsDevice; } }

        /// <summary>
        /// The sprite batch this control surface uses to draw it's surface.
        /// </summary>
        public SpriteBatch SpriteBatch { get { return mySpriteBatch; } }

        /// <summary>
        /// The control that this control surface is responsible for.
        /// </summary>
        public Control Control { get { return myControl; } }

        /// <summary>
        /// The full path to the texture to use as the background for this control, if this is null or empty then no texture will be used.
        /// </summary>
        public string BackgroundTexturePath
        {
            get { return myBackTexturePath; }
            set
            {
                if (value != myBackTexturePath)
                    isTextureReload = true;
                myBackTexturePath = value;
            }
        }

        /// <summary>
        /// The background color that will be used to clear the surface of the control.
        /// </summary>
        public Color BackgroundColor
        {
            get { return myBackColor; }
            set
            {
                if (myBackColor != value)
                    isRequireRedraw = true;
                myBackColor = value;
            }
        }

        /// <summary>
        /// Check if the control's surface texture should be redrawn, this value may only be set from inside this class or it's descendants.
        /// </summary>
        public bool IsRequireRedraw { get { return isRequireRedraw; } protected set { isRequireRedraw = value; } }

        /// <summary>
        /// Check if this control surface is visible.
        /// </summary>
        public bool IsVisible
        {
            get { return myControl.Visible && isVisible; }
            set
            {
                // If the control is visisble, this way visible, and is going to no longer be visible
                if (myControl.Visible && isVisible && !value)
                {
                    // Need to set the value and invalidate the surface of the control so that GDI takes over,
                    //  it must be done in this order since xna is drawing on a seperate thread from the form
                    //  and could potentially draw again before the invalidate goes through
                    isVisible = value;
                    try { myControl.Invalidate(false); }
                    catch { }
                }

                // Sucks setting this twice but it's better than holding a whole nother bool just to track
                //  the status
                isVisible = value;
            }
        }

        #endregion

        #region Constructor

        /// <summary>
        /// Create a new control surface.
        /// </summary>
        /// <param name="control">The control to link to the surface texture.</param>
        public ControlSurface(Game game, Control control)
        {
            myGame = game;
            myControl = control;
            myBackColor = Color.TransparentBlack;
            isRequireRedraw = false;
            isTextureReload = false;
            isVisible = true;
        }

        #endregion

        #region Initialization Methods

        /// <summary>
        /// Initialize the control surface, this will configure the viewport and hook events
        /// that need to be monitored to maintain the surface.
        /// </summary>
        public virtual void Initialize()
        {
            myViewport = GraphicsDevice.Viewport;
            myViewport.X = myControl.Location.X;
            myViewport.Y = myControl.Location.Y;
            myViewport.Width = myControl.Width;
            myViewport.Height = myControl.Height;

            myControl.Move += new System.EventHandler(OnControlMove);
            myControl.Resize += new System.EventHandler(OnControlResize);

            OnControlMove(null, null);
            OnControlResize(null, null);
        }

        #endregion

        #region Loading Methods

        /// <summary>
        /// Load the sprite batch, render target, and the background texture
        /// </summary>
        public virtual void LoadContent()
        {
            mySpriteBatch = new SpriteBatch(GraphicsDevice);

            myRenderTarget = new RenderTarget2D(GraphicsDevice,
                     myControl.Width, myControl.Height, 1,
                     GraphicsDevice.PresentationParameters.BackBufferFormat, MultiSampleType.None, 0);

            try { myBackTexture = Texture2D.FromFile(GraphicsDevice, myBackTexturePath); }
            catch { }

            isRequireRedraw = true;
            isTextureReload = false;
        }

        /// <summary>
        /// Unload the background texture and render target.
        /// </summary>
        public virtual void UnloadContent()
        {
            if (myRenderTarget != null)
            {
                if (!myRenderTarget.IsDisposed)
                    myRenderTarget.Dispose();

                myRenderTarget = null;
            }

            if (myBackTexture != null)
            {
                if (!myBackTexture.IsDisposed)
                    myBackTexture.Dispose();

                myBackTexture = null;
            }
        }

        #endregion

        #region Update Methods

        /// <summary>
        /// Perform any logic this control surface requires, this will also ensure that the render target is valid
        /// and the background texture is loaded.
        /// </summary>
        /// <param name="gameTime">Snapshot of the game time when Update was called.</param>
        public virtual void Update(GameTime gameTime)
        {
            // Ensure that the render target fits the current viewport
            if (myRenderTarget == null || myRenderTarget.Width < myViewport.Width || myRenderTarget.Height < myViewport.Height)
            {
                try
                {
                    myRenderTarget = new RenderTarget2D(GraphicsDevice,
                          myViewport.Width, myViewport.Height, 1,
                          GraphicsDevice.PresentationParameters.BackBufferFormat, MultiSampleType.None, 0);
                    isRequireRedraw = true;
                }
                catch { }
            }

            // Ensure that the background texture is up to date
            if (isTextureReload)
            {
                try
                {
                    myBackTexture = Texture2D.FromFile(GraphicsDevice, myBackTexturePath);
                    isRequireRedraw = true;
                    isTextureReload = false;
                }
                catch
                {
                    if (myBackTexture != null)
                    {
                        myBackTexture = null;
                        isRequireRedraw = true;
                        isTextureReload = false;
                    }
                }
            }
        }

        #endregion

        #region Drawing Methods

        /// <summary>
        /// Draw this control surface on to the back buffer.
        /// </summary>
        /// <param name="gameTime">Snapshot of the game time when Draw was called.</param>
        public void Draw(GameTime gameTime)
        {
            if (myTexture != null && IsVisible)
            {
                Viewport oldViewport = GraphicsDevice.Viewport;
                GraphicsDevice.Viewport = myViewport;
                GraphicsDevice.Clear(myBackColor);
                mySpriteBatch.Begin(SpriteBlendMode.AlphaBlend);

                // The co-ordinates in the destination of sprite batch calls are now relative in location since the
                //  viewport has been set, remember that !
                if (myBackTexture != null)
                    mySpriteBatch.Draw(myBackTexture, new Rectangle(0, 0, myViewport.Width, myViewport.Height), Color.White);
                mySpriteBatch.Draw(myTexture, new Rectangle(0, 0, myTexture.Width, myTexture.Height), Color.White);

                mySpriteBatch.End();

                if (myControl != null && !myControl.Disposing)
                {
                    GraphicsDevice.Present(
                        new Rectangle(myViewport.X, myViewport.Y, myViewport.Width, myViewport.Height),
                        new Rectangle(0, 0, myViewport.Width, myViewport.Height),
                        myControl.Handle);
                }
                GraphicsDevice.Viewport = oldViewport;
            }
        }

        /// <summary>
        /// Draw this control surface to it's internal texture.
        /// </summary>
        public void DrawToTexture()
        {
            // Only draw the surface texture if this control surface is visible and requires a redraw
            if (IsVisible)
            {
                DrawChildTextures();

                if (isRequireRedraw)
                {
                    // Store the old render target, swap in this control's render target, clear to the background
                    //  color, then draw the background image (if there is one) and allow the surface to be drawn
                    RenderTarget2D pRenderTarget = (RenderTarget2D)GraphicsDevice.GetRenderTarget(0);
                    GraphicsDevice.SetRenderTarget(0, myRenderTarget);
                    GraphicsDevice.Clear(Color.TransparentBlack);

                    DrawSurface();

                    // Send the surface to the control then restore the old render target and retrieve the texture
                    GraphicsDevice.SetRenderTarget(0, pRenderTarget);
                    myTexture = myRenderTarget.GetTexture();

                    isRequireRedraw = false;
                }
            }
        }

        /// <summary>
        /// Draw any textures child components of this control surface require before the main draw phase.
        /// </summary>
        protected virtual void DrawChildTextures() { }

        /// <summary>
        /// Draw on to the texture of this control surface.
        /// </summary>
        protected virtual void DrawSurface() { }

        #endregion

        #region Event Handlers

        /// <summary>
        /// When the control moves the viewport needs to be updated with the new location.
        /// </summary>
        private void OnControlMove(object sender, System.EventArgs e)
        {
            // The reason for getting the screen co-ordinates here is so that the location will
            //  match up on the back buffer
            System.Drawing.Point screenLocation = myControl.PointToScreen(myControl.Location);
            System.Drawing.Point formScreenLocation = myControl.FindForm().PointToScreen(System.Drawing.Point.Empty);

            myViewport.X = screenLocation.X - formScreenLocation.X;
            myViewport.Y = screenLocation.Y - formScreenLocation.Y;
        }

        /// <summary>
        /// When the control is resized the viewport needs to be updated with the new size.
        /// </summary>
        private void OnControlResize(object sender, System.EventArgs e)
        {
            myViewport.Width = myControl.Width;
            myViewport.Height = myControl.Height;
        }

        #endregion
    }
}
